import ssl
import aiohttp
import asyncio


# if sys.version_info >= (3, 5):
#     EventLoopType = t.Union[asyncio.BaseEventLoop, asyncio.AbstractEventLoop]
# else:
#     EventLoopType = asyncio.AbstractEventLoop


def get_or_create_event_loop() -> asyncio.AbstractEventLoop:
    try:
        loop = asyncio.get_event_loop()
        return loop
    except (RuntimeError, AssertionError):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop


# noinspection PyUnresolvedReferences
class TCPConnectorMixIn:

    # noinspection PyUnresolvedReferences
    def get_tcp_connector(self) -> aiohttp.TCPConnector:
        if self._connector_owner:
            # return valid connector
            if self._tcp_connector and not self._tcp_connector.closed:
                return self._tcp_connector
            # create ssl context if no valid connector is present
            ssl_context = ssl.create_default_context(cafile=self.cafile)

            # memoize tcp_connector for reuse
            # noinspection PyAttributeOutsideInit
            self._tcp_connector = aiohttp.TCPConnector(
                loop=self.loop,
                ssl_context=ssl_context,
                keepalive_timeout=self.keepalive_timeout,
            )
            return self._tcp_connector

        return self._tcp_connector

    def __del__(self):
        """
        Properly close owned connector on exit
        :return:
        """
        if self._connector_owner:
            connector = self.get_tcp_connector()
            not connector.closed and connector.close()




import typing as t

from rest_framework import status as drf_status


class FetchResult:
    """
    It replaces this:
        FetchResult = namedtuple('FetchResult', ['headers', 'result', 'status'])
    """
    def __init__(self, 
                 headers: t.Optional[dict]=None, 
                 result: t.Optional[t.Any]=None,
                 status: t.Optional[int]=None):
        self.headers = headers
        self._status = status
        self._result = result or {}

    def __repr__(self):
        return '<FetchResult: status={0.status}, headers={0.headers}, result={0.result}>'.format(self)

    def __bool__(self):
        return self.is_success()

    __nonzero__ = __bool__

    def __eq__(self, other: 'FetchResult'):
        return self.status == other.status and \
               self.headers == self.headers and \
               self.result == self.result

    def is_jsonrpc(self):
        return isinstance(self._result, dict) and \
               'jsonrpc' in self._result

    def is_success(self):
        return drf_status.is_success(self.status)

    def is_client_error(self):
        return drf_status.is_client_error(self.status)

    def is_informational(self):
        return drf_status.is_informational(self.status)

    def is_redirect(self):
        return drf_status.is_redirect(self.status)

    def is_server_error(self):
        return drf_status.is_server_error(self.status)

    @property
    def status(self):
        if self.is_jsonrpc() and 'error' in self._result:
            return self._result['error']['code']
        return self._status

    @property
    def result(self):
        return self._result



import asyncio
import json

from aiohttp import web
from aiohttp.web_request import Request


# noinspection PyUnusedLocal
async def home(request: Request):
    return web.Response(text='Home')


async def request_info(request: Request):
    bundle = {
        'cookies': dict(request.cookies.items()),
        'headers': dict(request.headers.items()),
        'query': dict(request.query.items()),
        'query_string': request.query_string,
        'charset': request.charset,
        'method': request.method
    }
    b_content = await request.read()
    try:
        encoding = request.charset or 'utf-8'
        t_content = b_content.decode(encoding)
    except UnicodeDecodeError:
        t_content = None

    if t_content is None:
        content = 'binary'
    else:
        try:
            content = json.loads(t_content)
        except json.JSONDecodeError:
            content = t_content

    bundle['content'] = content
    return web.json_response(bundle)


async def sleep_time(request: Request):
    st = float(request.match_info['sleep'])
    await asyncio.sleep(st)
    return web.json_response({'slept': st})


# noinspection PyUnusedLocal
async def view502(request: Request):
    message = 'GTFO. Gateway Took Fantastic Outing.'
    return web.Response(text=message, status=502)


def run_server():
    app = web.Application()
    app.router.add_get('/', home)
    app.router.add_route('*', '/request-info', request_info)
    app.router.add_route('*', '/502', view502)

    sleep_time_resource = app.router.add_resource('/sleep/{sleep:\d+}')
    sleep_time_resource.add_route('*', sleep_time)

    web.run_app(app, port=21571)


if __name__ == '__main__':
    run_server()



import pytest

from async_fetcher.fetch import FetchResult

pytestmark = pytest.mark.fetch_result

JSONRPC_RESULT_OK = {
    'id': 'some-unique-id',
    'jsonrpc': '2.0',
    'result': {'pong': 'ok'},
}
JSONRPC_RESULT_FAIL = {
    'id': 'some-unique-id',
    'jsonrpc': '2.0',
    'error': {
        'code': 500,
        'data': None,
        'message': 'OtherError: "some exception detail"',
        'name': 'OtherError'
    },
}


def test_init():
    fr = FetchResult(status=200, result={'res': 'res'}, headers={'headers': 'headers'})

    assert fr.status == 200
    assert fr.result == {'res': 'res'}
    assert fr.headers == {'headers': 'headers'}


def test_repr():
    fr = FetchResult(status=200, result='result', headers={'header': 'test'})
    assert fr.__repr__() == '<FetchResult: status=200, headers={\'header\': \'test\'}, result=result>'


def test_bool_expression():
    fr1 = FetchResult(status=200)
    assert fr1

    fr2 = FetchResult(status=500)
    assert not fr2

    fr3 = FetchResult(status=200, result=JSONRPC_RESULT_FAIL)
    assert not fr3

    fr3 = FetchResult(status=500, result=JSONRPC_RESULT_FAIL)
    assert not fr3


def test_is_jsonrpc():
    fr = FetchResult(status=200, result=JSONRPC_RESULT_OK)
    assert fr.is_jsonrpc()


def test_jsonrpc_fail_get_status_code():
    fr = FetchResult(status=200, result=JSONRPC_RESULT_FAIL)

    assert fr.is_jsonrpc()
    assert fr.status == 500


def test_result_is_not_iterable():
    fr = FetchResult(status=200, result=1)
    assert not fr.is_jsonrpc()

    # make it safer =)
    fr = FetchResult(status=200, result='')
    assert not fr.is_jsonrpc()

    fr = FetchResult(status=200, result=())
    assert not fr.is_jsonrpc()



import os
import time
import pytest

from async_fetcher.exceptions import AsyncFetchNetworkError
from async_fetcher.fetch import AsyncFetch, FetchResult

try:
    from aiohttp.client_exceptions import TimeoutError
except ImportError:  # aiohttp > 2.2.5
    from asyncio import TimeoutError

pytestmark = pytest.mark.fetcher

TEST_SERVER_URL = 'http://127.0.0.1:21571'


def build_url(*path_parts: str) -> str:
    return os.path.join(TEST_SERVER_URL, *path_parts)


# noinspection PyMethodMayBeStatic,PyShadowingNames
class FetcherTest:
    def test_mk_task(self):
        # request_info view in server.py
        url = build_url('request-info')
        tm = AsyncFetch.mk_task(url)

        assert tm == {
            'url': url, 'method': 'get', 'response_type': 'json', 'data': {},
            'timeout': None, 'do_not_wait': False, 'headers': {},
            'num_retries': -1, 'fail_silently': False
        }

        tm = AsyncFetch.mk_task(url, data={'test': 1}, method='post', headers={'X-LOL': 1})
        assert tm == {
            'url': url, 'data': '{"test": 1}', 'headers': {'X-LOL': 1, 'content-type': 'application/json'},
            'timeout': None, 'response_type': 'json', 'do_not_wait': False, 'method': 'post',
            'num_retries': -1, 'fail_silently': False
        }

    def test__mk_task__api_key_modifies_headers(self):
        url = build_url('request-info')
        tm = AsyncFetch.mk_task(url, api_key='api!key')
        assert tm['headers'] == {'api-key': 'api!key'}

    def test__mk_task__query_modifies_bundle_url(self):
        url = build_url('request-info')
        tm = AsyncFetch.mk_task(url, query={'key': True})
        assert tm['url'] == 'http://127.0.0.1:21571/request-info?key=True'

    def test__mk_task__language_code_modifies_headers(self):
        url = build_url('request-info')
        tm = AsyncFetch.mk_task(url, language_code='lol', headers={'X-LOL': 666})
        assert tm['headers'] == {'X-LOL': 666, 'accept-language': 'lol'}

    def test_fetch(self):
        url = build_url('request-info')
        task_bundle = AsyncFetch.mk_task(url, data='lol')
        af = AsyncFetch({})

        with af.get_client_session() as session:
            task = af.fetch(session, task_bundle)
            response = af.loop.run_until_complete(task)

            assert type(response) == FetchResult
            assert response.status == 200
            assert response.result['content'] == 'lol'

    def test_fetch__do_not_wait_flag_return_empty_response(self):
        url = build_url('error_url')
        task_bundle = AsyncFetch.mk_task(url, data='lol', do_not_wait=True)
        af = AsyncFetch({})

        with af.get_client_session() as session:
            task = af.fetch(session, task_bundle)
            response = af.loop.run_until_complete(task)

            assert type(response) == FetchResult
            assert response.status == 0
            assert response.result == {}
            assert response.headers is None

    def test_go(self):
        af = AsyncFetch({
            'first': AsyncFetch.mk_task(build_url('request-info')),
            'second': AsyncFetch.mk_task(build_url('request-info')),
            'fail': AsyncFetch.mk_task(build_url('404'))
        })
        responses = af.go()
        assert responses['fail'].status == 404
        assert responses['first'].status == 200
        assert responses['second'].status == 200

    def test_external_tcp_connector_alive(self):
        # check `AsyncFetch._connector_owner` flag set properly
        af0 = AsyncFetch({})
        tcp_connector = af0.get_tcp_connector()

        def af1__del__isolator():
            af1 = AsyncFetch({
                '1': AsyncFetch.mk_task(build_url('request-info')),
                '2': AsyncFetch.mk_task(build_url('request-info')),
            }, tcp_connector=tcp_connector)
            responses = af1.go()
            assert responses['1'].status == 200

        af1__del__isolator()
        assert tcp_connector.closed is False

    def test_tcp_connector_closed_on__del__(self):
        af0 = AsyncFetch({})
        tcp_connector = af0.get_tcp_connector()
        af0.__del__()
        assert tcp_connector.closed is True

    def test_connection_error(self):
        url = 'http://unknown'
        af = AsyncFetch({'0': AsyncFetch.mk_task(url)})

        with pytest.raises(AsyncFetchNetworkError):
            af.go()

    def test_502(self):
        url = build_url('502')
        af = AsyncFetch({i: AsyncFetch.mk_task(url) for i in range(10)}, num_retries=2, retry_timeout=0.1)

        with pytest.raises(AsyncFetchNetworkError):
            af.go()

    def test_really_async(self):
        task_map = {}

        def f():
            time1 = time.time()
            for i in range(10):
                task_map[len(task_map)] = AsyncFetch.mk_task(build_url('sleep', '0'))
                task_map[len(task_map)] = AsyncFetch.mk_task(build_url('sleep', '1'))

            af = AsyncFetch(task_map, num_retries=10, retry_timeout=0.1)
            af.go()
            time2 = time.time()
            return time2 - time1

        execution_time = f()
        assert execution_time < 2

    def test_fetch_raises_network_error(self):
        url = build_url('502')
        task_bundle = AsyncFetch.mk_task(url)
        af = AsyncFetch({}, num_retries=2, timeout=1, retry_timeout=1)

        with pytest.raises(AsyncFetchNetworkError):
            with af.get_client_session() as session:
                task = af.fetch(session, task_bundle)
                af.loop.run_until_complete(task)

    def test_fetch_handles_timeout_error(self):
        url = build_url('sleep', '10')
        task_bundle = AsyncFetch.mk_task(url)
        af = AsyncFetch({}, num_retries=1, timeout=1, retry_timeout=1)

        try:
            with af.get_client_session() as session:
                task = af.fetch(session, task_bundle)
                af.loop.run_until_complete(task)
        except Exception as e:
            assert isinstance(e, AsyncFetchNetworkError)
            assert isinstance(e.original_exception, TimeoutError)

    def test_retry_only_one_url(self):
        task_map = {
            0: AsyncFetch.mk_task(build_url('sleep', '1')),
            1: AsyncFetch.mk_task(build_url('502'), fail_silently=True)
        }
        af = AsyncFetch(task_map, num_retries=10, retry_timeout=0.2)
        res = af.go()
        assert res[1] == FetchResult(None, None, 0)



import typing as t
import os


def fake_ugettext_lazy(raw_str: str):
    return raw_str


# check if we have an active django environment (not just installed django package)
if os.getenv('DJANGO_SETTINGS_MODULE'):
    from django.utils.translation import ugettext_lazy as _
else:  # django package is not installed
    _ = fake_ugettext_lazy

try:
    from rest_framework.exceptions import APIException
except Exception as e:  # django.core.exceptions.ImproperlyConfigured
    if e.__class__.__name__ in ('ImproperlyConfigured', 'ImportError'):
        class APIException(Exception):
            default_detail = _('API error occurred.')
            default_code = 'error'

            # noinspection PyUnusedLocal
            def __init__(self, detail=None, code=None):
                self.detail = ''

            def __str__(self):
                return self.detail
    else:
        raise


class AsyncFetchError(APIException):
    def get_template(self):
        raise NotImplementedError()


class AsyncFetchReceiveError(AsyncFetchError):
    def __init__(self, detail=None, code=None,
                 service_name: str = '',
                 original_exception: t.Union[Exception, None] = None):
        self.original_exception = original_exception
        self.detail = str(self.get_template().format(service_name, str(original_exception)))

    def get_template(self):
        return _('Failed to receive data from `{0}` service. Original exception: `{1}`.')


class AsyncFetchNetworkError(AsyncFetchError):
    def __init__(self, detail=None, code=None,
                 service_name: str = '',
                 url: str = '',
                 original_exception: t.Union[Exception, None] = None,
                 retries_left: int = 0,
                 max_retries: int = 0,
                 response_code: int = 500):
        self.original_exception = original_exception
        self.code = code
        self.url = url
        self.retries_left = retries_left
        self.max_retries = max_retries
        self.response_code = response_code

        self.detail = str(self.get_template().format(
            str(service_name),
            str(url),
            str(retries_left),
            str(max_retries),
            str(response_code),
            str(original_exception) or type(original_exception)
        ))

    def get_template(self):
        return _('Network issue while requesting `{0}` service data from url `{1}` [{2} of {3} left]. '
                 'Last response code: {4}. Original exception: `{5}`.')
